/*
  The code in this file is licensed under Public Domain.
  The license can be found at http://creativecommons.org/licenses/publicdomain/.
  */

#include "AccountClass.h"

#include "Utilities.h"
#include "Logger.h"
#include <iostream>

AccountClass::AccountClass(AccountType type, QString guid, QDate oldestMonth, QDate selectedMonth)
{
    this->type = type;
    this->guid = guid;
    this->oldestMonth = oldestMonth;
    this->selectedMonth = selectedMonth;

    // Populate the isCalculated field
    for(int i = 0; i < DateToIndex(selectedMonth); i++)
    {
        isCalculated.append(false);
        totalPlanned.append(0);
        totalAccumulated.append(0);
    }


    if(type == ACCOUNT_TYPE_TOTAL_INCOME || type == ACCOUNT_TYPE_TOTAL_EXPENSE)
    {
        this->accountName = guid;
    }
    else if(type == ACCOUNT_TYPE_INCOME || type == ACCOUNT_TYPE_EXPENSE)
    {

        // Get the account name
        QSqlQuery query(QString("SELECT name, parent_guid FROM accounts WHERE guid='%1';")
                        .arg(guid));
        SqlQueryVerify(query, QString("Failed querying for the account name and parent_guid"));
        if(!query.next())
        {
            std::cerr << "No records returned for account name query. "
                    << " guid:" << qPrintable(guid)
                    << " query:" << qPrintable(query.lastQuery())
                    << " error:" << qPrintable(query.lastError().text()) << std::endl;
        }
        else
        {
            int nameIndex = query.record().indexOf("name");
            int parentIndex = query.record().indexOf("parent_guid");
            QString parent =  GetParentName(query.value(parentIndex).toString());
            QString name =  query.value(nameIndex).toString();
            if(!(parent == QString() || parent == QString("Expenses") || parent == QString("Income")))
            {
                name = parent + ":" + name;
            }

            this->accountName = name;
        }

        // Populate the account with the information given
        // Lookup all transactions from oldestDate to selectedDate
        QDateTime from, to;
        from.setDate(QDate(oldestMonth.year(), oldestMonth.month(), 1));
        int months = 1;
        while(from.date() < selectedMonth)
        {
            to = from;
            //to = to.addMonths(months - 1);
            to.setDate(SetToLastDayOfMonth(to.date()));

            QString str = QString(
                        "SELECT post_date, value_num FROM transactions, splits "
                        "WHERE splits.account_guid='%1' "
                        "AND splits.tx_guid=transactions.guid "
                        "AND transactions.post_date <= %2 "
                        "AND transactions.post_date >= %3;")
                    .arg(guid)
                    .arg(to.toString(TimeDateFormat))
                    .arg(from.toString(TimeDateFormat));
            //if(accountName == QString("Groceries"))
            //    std::cerr << "Name:" << qPrintable(accountName) << " Query:" << qPrintable(str) << std::endl;

            float amount = 0;

            QSqlQuery query;
            query.exec(str);
            SqlQueryVerify(query, QString("Error looking up transactions."));
            int amountIndex = query.record().indexOf("value_num");
            float nextAmount;
            while(query.next())
            {
                nextAmount = query.value(amountIndex).toString().toFloat();
                if(type == ACCOUNT_TYPE_INCOME)
                    nextAmount *= -1;
                nextAmount /= 100.0;
                amount += nextAmount;
            }

            QString monthKeyStr = from.toString(DateFormat);
            long monthKey = monthKeyStr.toLong();
            Q_ASSERT(monthKey != 0);

            // Put the data into the datastructure
            if(amount != 0.0)
            {
                monthlyTotals[monthKey] = amount;
            }

            // Create the Rule for this month
            RuleClass *pRule = new RuleClass(guid, from.date());
            this->monthlyRules[monthKey] = pRule;

            months++;
            from = from.addMonths(1);
        }
    }

}

AccountType AccountClass::getType(void)
{
    return this->type;
}

int AccountClass::DateToIndex(QDate date)
{
    // Compute the number of years between the oldest and selected dates
    int numYears = date.year() - this->oldestMonth.year();
    int numMonths = date.month() - this->oldestMonth.month();

    // Add in the number of years
    numMonths += numYears * 12;

    return numMonths;
}

bool AccountClass::getIsCalculated(QDate date)
{
    Q_ASSERT(date.isValid() == true);

    LogMsg("AccountClass/get", LOG_DEBUG,
           "Account:%s getIsCalculated for month:%s.",
           qPrintable(getName()), qPrintable(date.toString()));
    int index = DateToIndex(date);
    return isCalculated.at(index);
}

void AccountClass::setIsCalculated(bool calc, QDate date)
{
    int index = DateToIndex(date);
    isCalculated[index] = calc;
}

void AccountClass::setTotalAccountTotal(float total, QDate date)
{
    if(type == ACCOUNT_TYPE_TOTAL_INCOME || type == ACCOUNT_TYPE_TOTAL_EXPENSE)
    {
        int index = DateToIndex(date);
        totalPlanned[index] = total;
    }
    else
    {
        LogMsg("AccountClass/get", LOG_ERROR,
               "ERROR: Attempting to use a 'Total' account for non-total account. Account:%s month:%s.",
               qPrintable(accountName), qPrintable(date.toString()));
    }
}

void AccountClass::setTotalAccountExtra(float extra, QDate date)
{
    if(type == ACCOUNT_TYPE_TOTAL_INCOME || type == ACCOUNT_TYPE_TOTAL_EXPENSE)
    {
        int index = DateToIndex(date);
        totalAccumulated[index] = extra;
    }
    else
    {
        LogMsg("AccountClass/get", LOG_ERROR,
               "ERROR: Attempting to use a 'Total' account for non-total account. Account:%s month:%s.",
               qPrintable(accountName), qPrintable(date.toString()));
    }
}

float AccountClass::getTotalAccountTotal(QDate date)
{
    if(type == ACCOUNT_TYPE_TOTAL_INCOME || type == ACCOUNT_TYPE_TOTAL_EXPENSE)
    {
        int index = DateToIndex(date);
        return totalPlanned.at(index);
    }
    else
    {
        LogMsg("AccountClass/get", LOG_ERROR,
               "ERROR: Attempting to use a 'Total' account for non-total account. Account:%s month:%s.",
               qPrintable(accountName), qPrintable(date.toString()));
    }
    return 0;
}

float AccountClass::getTotalAccountExtra(QDate date)
{
    if(type == ACCOUNT_TYPE_TOTAL_INCOME || type == ACCOUNT_TYPE_TOTAL_EXPENSE)
    {
        int index = DateToIndex(date);
        return totalAccumulated.at(index);
    }
    else
    {
        LogMsg("AccountClass/get", LOG_ERROR,
               "ERROR: Attempting to use a 'Total' account for non-total account. Account:%s month:%s.",
               qPrintable(accountName), qPrintable(date.toString()));
    }
    return 0;
}


float AccountClass::getCalculatedPlanned(QDate date)
{
    if(type == ACCOUNT_TYPE_TOTAL_INCOME || type == ACCOUNT_TYPE_TOTAL_EXPENSE)
    {
        LogMsg("AccountClass/get", LOG_ERROR,
               "ERROR: Attempting to use a 'Total' account for non-total account. Account:%s month:%s.",
               qPrintable(accountName), qPrintable(date.toString()));
        return 0;
    }

    //std::cerr << "GetCalculatedPlanned: Name:" << qPrintable(accountName) << " planned:" << totalPlanned.at(DateToIndex(date)) << std::endl;
    return totalPlanned.at(DateToIndex(date));
}

float AccountClass::getCalculatedAccumulated(QDate date)
{
    if(type == ACCOUNT_TYPE_TOTAL_INCOME || type == ACCOUNT_TYPE_TOTAL_EXPENSE)
    {
        LogMsg("AccountClass/get", LOG_ERROR,
               "ERROR: Attempting to use a 'Total' account for non-total account. Account:%s month:%s.",
               qPrintable(accountName), qPrintable(date.toString()));
        return 0;
    }

    int index = DateToIndex(date);
    // Return zero for non-existant months to be able to sum all of the months
    if(index < 0)
        return 0;
    else
        return totalAccumulated.at(index);
}

void AccountClass::setCalculatedPlanned(float planned, QDate date)
{
    if(type == ACCOUNT_TYPE_TOTAL_INCOME || type == ACCOUNT_TYPE_TOTAL_EXPENSE)
    {
        LogMsg("AccountClass/get", LOG_ERROR,
               "ERROR: Attempting to use a 'Total' account for non-total account. Account:%s month:%s.",
               qPrintable(accountName), qPrintable(date.toString()));
        return;
    }

    if(accountName.toLower().indexOf(QString("tithing")) >= 0)
    {
        std::cerr << "";
    }
    int index = DateToIndex(date);

    Q_ASSERT(index < isCalculated.size());
    Q_ASSERT(index < totalPlanned.size());

    isCalculated[index] = true;
    totalPlanned[index] = planned;
}

void AccountClass::setCalculatedAccumulated(float accumulated, QDate date)
{
    if(type == ACCOUNT_TYPE_TOTAL_INCOME || type == ACCOUNT_TYPE_TOTAL_EXPENSE)
    {
        LogMsg("AccountClass/get", LOG_ERROR,
               "ERROR: Attempting to use a 'Total' account for non-total account. Account:%s month:%s.",
               qPrintable(accountName), qPrintable(date.toString()));
        return;
    }

    // Get the last months and add it to this months
    int index = DateToIndex(date);
    LogMsg("AccountClass/setCalculatedAccumulated", LOG_DEBUG,
           "Set Accumulated: Account:%s Amount:%0.2f Date:%s Index:%d",
           qPrintable(this->accountName), accumulated, qPrintable(date.toString()), index);
    if(index == -1)
    {
        return;
    }
    if(index == 0)
    {
        totalAccumulated[index] = accumulated;
    }
    else
    {
        totalAccumulated[index] = totalAccumulated[index-1] + accumulated;
    }
}

RuleClass* AccountClass::GetMonthlyRule(QDate date)
{
    if(type == ACCOUNT_TYPE_TOTAL_INCOME || type == ACCOUNT_TYPE_TOTAL_EXPENSE)
    {
        LogMsg("AccountClass/get", LOG_ERROR,
               "ERROR: Attempting to use a 'Total' account for non-total account. Account:%s month:%s.",
               qPrintable(accountName), qPrintable(date.toString()));
        return NULL;
    }

    QDateTime month;
    month.setTime(QTime());
    month.setDate(QDate(date.year(), date.month(), 1));

    RuleClass *pRule = monthlyRules.value(month.toString(DateFormat).toLong());

    // Total Income and Expenses have no rules
    if(pRule == NULL)
        return NULL;

    return pRule;
}

float AccountClass::GetMonthlyTotal(QDate date)
{
    if(type == ACCOUNT_TYPE_TOTAL_INCOME || type == ACCOUNT_TYPE_TOTAL_EXPENSE)
    {
        LogMsg("AccountClass/get", LOG_ERROR,
               "ERROR: Attempting to use a 'Total' account for non-total account. Account:%s month:%s.",
               qPrintable(accountName), qPrintable(date.toString()));
        return 0;
    }

    QDateTime month;
    month.setTime(QTime());
    month.setDate(QDate(date.year(), date.month(), 1));
    float total;
    long key = month.toString(DateFormat).toLong();
    total = monthlyTotals.value(key);
    return total;
}

QString AccountClass::getName(void)
{
    return this->accountName;
}

QString AccountClass::GetParentName(QString parent_guid)
{
    if(type == ACCOUNT_TYPE_TOTAL_INCOME || type == ACCOUNT_TYPE_TOTAL_EXPENSE)
    {
        LogMsg("AccountClass/get", LOG_ERROR,
               "ERROR: Attempting to use a 'Total' account for non-total account. Account:%s",
               qPrintable(accountName));
        return 0;
    }

    QSqlQuery query(
            QString("SELECT * FROM accounts WHERE guid=\"%1\";").arg(parent_guid));

    if(!query.next())
    {
        return QString();
    }

    int nameIndex = query.record().indexOf("name");
    int parentIndex = query.record().indexOf("parent_guid");

    QString new_parent_guid = query.value(parentIndex).toString();
    QString name = query.value(nameIndex).toString();

    QString parent = GetParentName(new_parent_guid);

    if(parent == QString() || parent == QString("Expenses") || parent == QString("Income"))
        return name;
    else
        return parent + ":" + name;
}
