/*
  The code in this file is licensed under Public Domain.
  The license can be found at http://creativecommons.org/licenses/publicdomain/.
  */

#include "BudgetClass.h"
#include "Utilities.h"
#include "Logger.h"

#include <QtSql>
#include <QMessageBox>
#include <QFileDialog>

#include <iostream>

BudgetClass::BudgetClass(QDate debugCurrentMonth)
{
    CreateDatabaseConnection();

    // Only continue if there is a database open
    if(databaseOpen == false)
        exit(-1);

    selectedMonth = debugCurrentMonth;
    LogMsg("BudgetClass/start", LOG_INFO, "Newest Date = %s", qPrintable(selectedMonth.toString()));

    QSqlQuery query, query2;

    // Create the budget rules table if it doesn't exist
    query.exec("SELECT * FROM budget_rules;");
    if(query.lastError().isValid()
        && query.lastError().text().contains("no such table"))
    {
        // budget_rules doesn't exist.  Create it.
        query.exec("CREATE TABLE budget_rules ("
                               "guid text(32) PRIMARY KEY NOT NULL, "
                               "account_guid text(32) NOT NULL, "
                               "month text(14) NOT NULL, "
                               "accountMax text(32) NOT NULL, "
                               "monthlyMax text(32) NOT NULL, "
                               "monthlyFixed text(32) NOT NULL, "
                               "percentage text(32) NOT NULL);");
        SqlQueryVerify(query, "Error creating the budget_rules table.");
    }

    // Find the oldest date
    query.exec("SELECT post_date FROM transactions ORDER BY post_date ASC;");
    SqlQueryVerify(query, "Error trying to find the oldest transaction.");

    if(query.next())
    {
        QDateTime old;
        old = old.fromString(query.value(query.record().indexOf("post_date")).toString(),
                       TimeDateFormat);

        //std::cerr << "Oldest Date:" << qPrintable(query.value(query.record().indexOf("post_date")).toString())
        //        << " Date:" << qPrintable(old.date().toString()) << std::endl;
        oldestMonth = old.date();
        LogMsg("BudgetClass/start", LOG_INFO, "OldestDate = %s", qPrintable(oldestMonth.toString()));
    }

    // Create the totaling accounts
    pTotalIncome = new AccountClass(ACCOUNT_TYPE_TOTAL_INCOME, TOTAL_INCOME, oldestMonth, selectedMonth);
    pTotalExpenses = new AccountClass(ACCOUNT_TYPE_TOTAL_EXPENSE, TOTAL_EXPENSE, oldestMonth, selectedMonth);
    pHeader = new AccountClass(ACCOUNT_TYPE_HEADER, "", QDate(), QDate());

    // Header row
    rowMapping += QString();
    rowMapping += QString(TOTAL_INCOME);

    // Get all of the income guids and create accounts for each
    query.exec("SELECT * FROM accounts WHERE account_type='INCOME'");
    SqlQueryVerify(query, "Error querying all of the income accounts.");
    int guidIndex = query.record().indexOf("guid");
    while(query.next())
    {
        QString guid = query.value(guidIndex).toString();

        // Only accept income accounts that don't have children.
        // Only accept leaf accounts.
        query2.exec(QString("SELECT * FROM accounts WHERE parent_guid='%1';").arg(guid));
        SqlQueryVerify(query, "Error querying the income account to see it's children.");
        if(!query2.next())
        {
            incomeAccounts[guid] = new AccountClass(ACCOUNT_TYPE_INCOME, guid, oldestMonth, selectedMonth);
            rowMapping += guid;
        }
    }

    rowMapping += QString(TOTAL_EXPENSE);

    // Get all of the expense guids and create accounts for each
    query.exec("SELECT * FROM accounts WHERE account_type='EXPENSE'");
    SqlQueryVerify(query, "Error querying all of the expense accounts.");
    guidIndex = query.record().indexOf("guid");
    while(query.next())
    {
        QString guid = query.value(guidIndex).toString();

        // Only accept expense accounts that don't have children.
        // Only accept leaf accounts.
        query2.exec(QString("SELECT * FROM accounts WHERE parent_guid='%1';").arg(guid));
        SqlQueryVerify(query, "Error querying the expense account to see it's children.");
        if(!query2.next())
        {
            expenseAccounts[guid] = new AccountClass(ACCOUNT_TYPE_EXPENSE, guid, oldestMonth, selectedMonth);
            rowMapping += guid;
        }
    }


}

AccountClass* BudgetClass::getAccount(QString guid) const
{
    if(guid == QString(TOTAL_INCOME))
        return pTotalIncome;
    if(guid == QString(TOTAL_EXPENSE))
        return pTotalExpenses;


    AccountClass *pAccount;
    pAccount = incomeAccounts.value(guid);
    if(pAccount == NULL)
    {
        pAccount = expenseAccounts.value(guid);
    }
    Q_ASSERT(pAccount != NULL);

    return pAccount;
}

RuleClass* BudgetClass::GetRule(QString guid, QDate month)
{
    Q_ASSERT(month.isValid() == true);
    //std::cerr << "Date: " << qPrintable(month.toString()) << std::endl;

    AccountClass *pAccount = getAccount(guid);

    if(pAccount == NULL)
        return NULL;


    RuleClass *pRule = pAccount->GetMonthlyRule(month);

    // Total Expenses and Income have no rules
    if(pRule == NULL)
        return NULL;

    LogMsg("BudgetClass/rule", LOG_DEBUG, "Requesting rule for:%s :%s",
           qPrintable(pAccount->getName()),qPrintable(month.toString()));

    // Ensure that they are identical
    if(pRule->getDate() != month)
    {
        LogMsg("BudgetClass/rule", LOG_ERROR,
               "Rule found date:%s doesn't match requested date:%s",
               qPrintable(pRule->getDate().toString()), qPrintable(month.toString()));
    }
    return pRule;
}

void BudgetClass::ComputeMonthlyTotal(QString guid, QDate month, float &actual, float &planned, float &accumulated)
{
    actual = 0;
    planned = 0;
    accumulated = 0;

    AccountClass *pAccount;
    if(guid == QString(TOTAL_INCOME))
    {
        pAccount = pTotalIncome;
    }
    else if(guid == QString(TOTAL_EXPENSE))
        pAccount = pTotalExpenses;
    else
    {
        // See if this is in the income accounts
        pAccount = incomeAccounts.value(guid);
        if(pAccount == NULL)
        {
            // It is assumed to be here
            pAccount = expenseAccounts.value(guid);
        }
    }

    Q_ASSERT(pAccount != NULL);

    if(!pAccount->getIsCalculated(month))
    {
        LogMsg("BudgetClass/request", LOG_INFO, "Need to compute budget for account:%s date:%s",
               qPrintable(pAccount->getName()), qPrintable(month.toString()));

        ComputeBudget(month);
    }

    actual =  pAccount->GetMonthlyTotal(month);
    planned = pAccount->getCalculatedPlanned(month);
    accumulated = pAccount->getCalculatedAccumulated(month);
    LogMsg("BudgetClass/request", LOG_DEBUG, "Returning request for account:%s actual:%0.2f planned:%0.2f accumulated:%0.2f",
           qPrintable(pAccount->getName()), actual, planned, accumulated);
}

void BudgetClass::ComputeBudget(QDate month)
{
    float IncomeTotal = 0, ExpenseTotal = 0;
    AccountClass *pAccount;

    LogMsg("BudgetClass/compute", LOG_DEBUG, "Computing budget for date: %s",
       qPrintable(month.toString()));

    // Sum the income for the month
    {
        QHashIterator<QString, AccountClass*> i(incomeAccounts);
        while (i.hasNext()) {
            i.next();
            pAccount = i.value();
            float monthlyAmount = pAccount->GetMonthlyTotal(month);
            IncomeTotal += monthlyAmount;
            pAccount->setCalculatedPlanned(monthlyAmount, month);

            LogMsg("BudgetClass/compute", LOG_DEBUG, "Income: Account:%s Planned:%0.2f",
               qPrintable(pAccount->getName()), monthlyAmount);
        }
    }

    LogMsg("BudgetClass/compute", LOG_DEBUG, "Total Income: %0.2f",
       IncomeTotal);

    // Store the Income amount
    pTotalIncome->setTotalAccountTotal(IncomeTotal, month);

    float TotalPlanned = 0;
    QList<AccountClass*> variableAccounts;  // Store the accounts that have variable planned amounts.

    // Get the total expenses for the month
    QHashIterator<QString, AccountClass*> i(expenseAccounts);
    while (i.hasNext()) {
        i.next();
        pAccount = i.value();

        // Is this a tithing expense account?
        if(pAccount->getName().toLower().indexOf(QString("tithing"), 0, Qt::CaseInsensitive) != -1)
        {
            // This is.  Update the tithing expense.
            float amount = IncomeTotal * 0.10;
            pAccount->setCalculatedPlanned(amount, month);
        }

        // Add the amount to the total expenses for the month
        ExpenseTotal += pAccount->GetMonthlyTotal(month);

        // Store all of the accounts that have/may have varying monthly planned amounts
        RuleClass *pRule = pAccount->GetMonthlyRule(month);
        if(pRule->getPercent() != 0 || pRule->getAccountMax() != 0)
        {
            // Because an account max can modify the remaining accounts (totalplanned),
            // put the limiting accounts first.
            if(pRule->getAccountMax() != 0)
                variableAccounts.append(pAccount);
            else
                variableAccounts.prepend(pAccount);

            // If the percentage is zero and the monthly fixed is non-zero add that to the planned amount
            if(pRule->getPercent() == 0)
            {
                Q_ASSERT(pRule->getMonthlyFixed() != 0);
                TotalPlanned += pRule->getMonthlyFixed();
            }
            LogMsg("BudgetClass/compute", LOG_DEBUG,
                   "Found variable account: %s Actual:%0.2f Sum:%0.2f",
                   qPrintable(pAccount->getName()), pAccount->GetMonthlyTotal(month), ExpenseTotal);
        }
        else
        {
            float plannedAmount;
            if(pRule->getMonthlyFixed() != 0)
            {
                // Use the Rules monthly fixed amount
                plannedAmount = pRule->getMonthlyFixed();
            }
            // If this is the tithing account, get the already calcuated planned Amount
            else if(pAccount->getName().toLower().indexOf(QString("tithing"), 0, Qt::CaseInsensitive) != -1)
            {
                plannedAmount = pAccount->getCalculatedPlanned(month);
            }
            else
            {
                // Accept whatever expenses have posted against the account
                plannedAmount = pAccount->GetMonthlyTotal(month);
            }

            pAccount->setCalculatedPlanned(plannedAmount, month);
            TotalPlanned += plannedAmount;

            LogMsg("BudgetClass/compute", LOG_DEBUG, "Expense Account: %s Total:%0.2f Planned:%0.2f",
                   qPrintable(pAccount->getName()), pAccount->GetMonthlyTotal(month), plannedAmount);
        }
    }

    // Store the total Expenses
    pTotalExpenses->setTotalAccountTotal(ExpenseTotal, month);

    LogMsg("BudgetClass/compute", LOG_DEBUG, "Total Income: %0.2f Total Expenses: %0.2f TotalPlanned: %0.2f Remaining=%0.2f",
           IncomeTotal, ExpenseTotal, TotalPlanned, (IncomeTotal - TotalPlanned));

    // Calculate the varying planned accounts
    if(variableAccounts.size() == 0)
    {
        pTotalIncome->setTotalAccountExtra(TotalPlanned, month);
    }
    else
    {
        bool doneCalculating =  false;
        while(doneCalculating == false && variableAccounts.size() > 0)
        {
            doneCalculating = true;

            // Compute the weighted average of the remaining accounts
            float totalPercentage = 0;
            for(int i = 0; i < variableAccounts.size(); i++)
            {
                AccountClass *pAccount = variableAccounts.at(i);
                RuleClass *pRule = pAccount->GetMonthlyRule(month);

                if(pRule->getPercent() > 0)
                {
                    totalPercentage += pRule->getPercent();
                }
            }

            float weightedAverage = 1;
            if(totalPercentage > 0)
                weightedAverage = 100.0 / totalPercentage;


            LogMsg("BudgetClass/compute", LOG_DEBUG, "Variable Accounts: Percent Total: %0.2f Weighted average:%f",
                   totalPercentage, weightedAverage);

            // Next start computing the remaining planned amounts
            {
                QMutableListIterator<AccountClass*> i(variableAccounts);
                while (i.hasNext())
                {

                    AccountClass *pAccount = i.next();
                    RuleClass *pRule = pAccount->GetMonthlyRule(month);
                    QString str;
                    str = "Account:" + pAccount->getName();

                    if(pAccount->getName() == QString("Auto:Gas"))
                    {
                        std::cerr << "";
                    }

                    // Compute the planned amount
                    float planned = pRule->getMonthlyFixed()
                                    + pRule->getPercent() / 100.0 * weightedAverage
                                                          * (IncomeTotal - TotalPlanned);
                                                          //* (pTotalIncome->getCalculatedTotalExpenses(month) - TotalPlanned);
                    str += " planned:" + QString::number(planned);

                    // Does this exceed the monthly maximum?
                    if(pRule->getMonthlyMax() > 0 && planned > pRule->getMonthlyMax())
                    {
                        // Set the planned amount to the monthly max
                        planned = pRule->getMonthlyMax();
                        pAccount->setCalculatedPlanned(planned, month);

                        // Remove this account from the list of accounts to calculate
                        i.remove();

                        str += " Monthly Max hit:" + QString::number(planned) + " New Total:" + QString::number(TotalPlanned);
                        LogMsg("BudgetClass/compute", LOG_DEBUG, "%s", qPrintable(str));

                        // This account hitting it's max messes up the weighted percentage.  Recalculate.
                        doneCalculating = false;
                    }

                    // Does this exceed the maximum for the account?
                    if(pRule->getAccountMax() > 0
                       && planned + pAccount->getCalculatedAccumulated(month.addMonths(-1)) - pAccount->GetMonthlyTotal(month) > pRule->getAccountMax())
                    {
                        planned = pRule->getAccountMax() - pAccount->getCalculatedAccumulated(month.addMonths(-1)) + pAccount->GetMonthlyTotal(month);
                        pAccount->setCalculatedPlanned(planned, month);

                        i.remove();


                        str += " Account Max hit:" + QString::number(planned) + " New Total:" + QString::number(TotalPlanned);
                        LogMsg("BudgetClass/compute", LOG_DEBUG, "%s", qPrintable(str));
                        // This messes up the weighted percentage.  Recalculate.
                        doneCalculating = false;
                    }

                    // If the doneCalculating flag is set to false, recalculate the weighted percentage.
                    // This break is here to allow the monthly max and the account max to be hit.
                    if(doneCalculating == false)
                    {
                        // Add the value to the total planned.
                        TotalPlanned += planned;
                        break;
                    }

                    // This value is fine. Continue;
                    pAccount->setCalculatedPlanned(planned, month);
                    LogMsg("BudgetClass/compute", LOG_DEBUG, "%s", qPrintable(str));
                }
            }
        }
    }

    // Set the accumulated amounts and compute the total planned
    {
        float totPlanned = 0;
        QHashIterator<QString, AccountClass*> i(expenseAccounts);
        while (i.hasNext()) {
            i.next();
            pAccount = i.value();

            float Extra = pAccount->getCalculatedPlanned(month) - pAccount->GetMonthlyTotal(month);
            pAccount->setCalculatedAccumulated(Extra, month);
            totPlanned += pAccount->getCalculatedPlanned(month);
            LogMsg("BudgetClass/compute", LOG_DEBUG, "Calculate tot: Account:%s Accum:%0.2f Planned:%0.2f",
                   qPrintable(pAccount->getName()), Extra, pAccount->getCalculatedPlanned(month));
        }

        // Next compute the difference between the income and the total planned
        // to determine if there is a difference.
        pTotalIncome->setTotalAccountExtra(IncomeTotal - totPlanned, month);
        pTotalExpenses->setTotalAccountExtra(totPlanned, month);
     }

}

bool BudgetClass::CreateDatabaseConnection(void)
{
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE");
    QString fileName;
    fileName = QFileDialog::getOpenFileName(
                 NULL,
                 QString("Open gnucash Database"),
                 QDir::currentPath(),
                 QString("Database Files (*.*)"));
    if(fileName.isEmpty() == true)
    {
        QMessageBox::warning(0, QObject::tr("File Error"),
                             QString("No file selected."));
        return false;
    }

    db.setDatabaseName(fileName);

    if (!db.open())
    {
        QMessageBox::warning(0, QObject::tr("Database Error"),
                             db.lastError().text());
        databaseOpen = false;
        return databaseOpen;
    }
    else
    {
        LogMsg("BudgetClass/db", LOG_INFO, "Database Opened. Name:%s", qPrintable(db.databaseName()));
    }

    // Query the database driver to determine supported/unsupported features.
    bool transactions = db.driver()->hasFeature(QSqlDriver::Transactions);
    bool querysize = db.driver()->hasFeature(QSqlDriver::QuerySize);
    bool BLOB = db.driver()->hasFeature(QSqlDriver::BLOB);
    bool PrepareQueryies = db.driver()->hasFeature(QSqlDriver::PreparedQueries);
    bool SimpleLocking = db.driver()->hasFeature(QSqlDriver::SimpleLocking);

    LogMsg("BudgetClass/db", LOG_DEBUG, "Database %s %s.",
           (transactions ? "supports" : "doesn't suport"), "Transactions");
    LogMsg("BudgetClass/db", LOG_DEBUG, "Database %s %s.",
           (querysize ? "supports" : "doesn't suport"), "querysize");
    LogMsg("BudgetClass/db", LOG_DEBUG, "Database %s %s.",
           (BLOB ? "supports" : "doesn't suport"), "BLOB");
    LogMsg("BudgetClass/db", LOG_DEBUG, "Database %s %s.",
           (PrepareQueryies ? "supports" : "doesn't suport"), "PrepareQueryies");
    LogMsg("BudgetClass/db", LOG_DEBUG, "Database %s %s.",
           (SimpleLocking ? "supports" : "doesn't suport"), "SimpleLocking");

    databaseOpen = true;
    return databaseOpen;
}

int BudgetClass::rowCount(void)
{
    return rowMapping.size();
}

int BudgetClass::columnCount(void)
{
    // Compute the number of years between the oldest and selected dates
    int numYears = this->selectedMonth.year() - this->oldestMonth.year();
    int numMonths = this->selectedMonth.month() - this->oldestMonth.month();

    // Add in the number of years
    numMonths += numYears * 12;

    // Two columns for each month
    return numMonths * 2;
}

QString BudgetClass::getGuid(int row)
{
    return rowMapping.at(row);
}

QDate BudgetClass::columnToDate(int column) const
{
    if(column % NUM_COLUMNS != 0)
        column -= column % NUM_COLUMNS;
    return oldestMonth.addMonths(column / NUM_COLUMNS);
}

QString BudgetClass::rowName(int row)
{
    QString guid = getGuid(row);

    if(    guid == QString(TOTAL_INCOME)
        || guid == QString(TOTAL_EXPENSE)
        || guid == QString())
        return guid;

    AccountClass *pAccount = incomeAccounts.value(guid);
    if(pAccount == NULL)
    {
        pAccount = expenseAccounts.value(guid);
    }

    Q_ASSERT(pAccount);

    return pAccount->getName();
}
