/*
  The code in this file is licensed under Public Domain.
  The license can be found at http://creativecommons.org/licenses/publicdomain/.
  */

#include "BudgetModel.h"
#include <QtSql>

#include <iostream>
#include <QMessageBox>

#include "Utilities.h"
#include <qglobal.h>



BudgetModel::BudgetModel(QObject *pParent) : QAbstractTableModel(pParent)
{
    pBudget = new BudgetClass(QDate(2009,1,1));
}


int BudgetModel::rowCount(const QModelIndex & /*parent*/ ) const
{
    return pBudget->rowCount();
}

int BudgetModel::columnCount(const QModelIndex &/*parent*/) const
{
    return pBudget->columnCount();
}

QVariant BudgetModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(role == Qt::DisplayRole)
    {
        if(orientation == Qt::Vertical)
        {
            int row = section;
            return QVariant(pBudget->rowName(row));
            //return QVariant(QString("%1").arg(row));
        }
        else
        {
            int column = section;
            return QVariant(pBudget->columnToDate(column).toString("MM/yy"));
            //return QVariant(QString("%1 Col").arg(column));
        }
    }
    else if(role == Qt::TextAlignmentRole)
    {
        if(orientation == Qt::Vertical)
        {
            int row = section;
            QString name = pBudget->rowName(row);
            if(name == QString(TOTAL_INCOME) || name == QString(TOTAL_EXPENSE))
                return QVariant(Qt::AlignLeft);
            else
                return QVariant(Qt::AlignRight);
        }
    }
    else if(role == Qt::UserRole)
    {
        if(orientation == Qt::Vertical)
        {
            int row = section;

            QVariant val;
            val.setValue(pBudget->getGuid(row));
            return val;
        }
    }


    return QVariant();
}

AccountClass* BudgetModel::getAccount(int row) const
{
    QString guid = headerData(row, Qt::Vertical, Qt::UserRole).toString();
    return pBudget->getAccount(guid);
}
QDate BudgetModel::getDate(int column) const
{
    return pBudget->columnToDate(column);
}

const BudgetClass* BudgetModel::getBudget(void) const
{
    return pBudget;
}

QVariant BudgetModel::data(const QModelIndex &index, int role) const
{
    QVariant val;
    if(role == Qt::DisplayRole)
    {
        // Header Row
        if(index.row() == 0)
        {
            if(index.column() % NUM_COLUMNS == RULE_COLUMN)
                return QVariant(QString("Rule"));
            if(index.column() % NUM_COLUMNS == ACTUAL_COLUMN)
                return QVariant(QString("Actual"));
            if(index.column() % NUM_COLUMNS == PLANNED_COLUMN)
                return QVariant(QString("Plan"));
            if(index.column() % NUM_COLUMNS == ACCUMULATED_COLUMN)
                return QVariant(QString("Accum"));
            return QVariant(QString("Unknown"));
        }

        val = this->headerData(index.row(), Qt::Vertical, Qt::UserRole);
        QString guid = val.toString();
        AccountClass *pAccount = pBudget->getAccount(guid);

        // The header should have been dealt with already
        Q_ASSERT(pAccount->getType() != ACCOUNT_TYPE_HEADER);

        // Deal with the total rows
        if(pAccount->getType() == ACCOUNT_TYPE_TOTAL_INCOME
           || pAccount->getType() == ACCOUNT_TYPE_TOTAL_EXPENSE)
        {
            if(index.column() % NUM_COLUMNS == ACTUAL_COLUMN)
            {
                QDate date = pBudget->columnToDate(index.column());
                float total = pAccount->getTotalAccountTotal(date);

                QString str;
                str.sprintf("%s", qPrintable(FormatMoney(total)));
                return QVariant(str);
            }

            if(index.column() % NUM_COLUMNS == PLANNED_COLUMN)
            {
                QDate date = pBudget->columnToDate(index.column());
                float extra = pAccount->getTotalAccountExtra(date);
                QString str;
                if(extra != 0)
                    str.sprintf("%s", qPrintable(FormatMoney(extra)));
                return QVariant(str);
            }

            // Otherwise return nothing
            return QVariant();
        }

        // Remaining rows are income rows and expense rows

        // Rule Column
        if(pAccount->getType() == ACCOUNT_TYPE_EXPENSE
           && index.column() % NUM_COLUMNS == RULE_COLUMN)
        {
            QDate date = pBudget->columnToDate(index.column());
            RuleClass *pRule = pBudget->GetRule(guid, date);

            Q_ASSERT(pRule != NULL);
            /*
            if(!pRule)
            {
                // No Rule
                return QVariant();
            }*/

            QVariant returnVal;
            returnVal.setValue(pRule);
            return returnVal;
        }

        float actual, planned, accumulated;
        actual = 0;
        planned = 0;
        accumulated = 0;
        QVariant returnVal;
        // Total Value
        pBudget->ComputeMonthlyTotal(val.toString(), pBudget->columnToDate(index.column()), actual, planned, accumulated);

        // Actual Column
        if(index.column() % NUM_COLUMNS == ACTUAL_COLUMN)
        {
            QString str;
            if(pAccount->getType() == ACCOUNT_TYPE_TOTAL_INCOME)
            {
                if(planned != 0)
                {
                    str.sprintf("(+%s) %s", qPrintable(FormatMoney(planned)), qPrintable(FormatMoney(accumulated)));
                }
                else
                {
                    str.sprintf("%s", qPrintable(FormatMoney(accumulated)));
                }
            }
            else if(pAccount->getType() == ACCOUNT_TYPE_TOTAL_EXPENSE)
            {
                str.sprintf("%s", qPrintable(FormatMoney(accumulated)));
            }
            else if(pAccount->getType() == ACCOUNT_TYPE_INCOME)
            {
                str.sprintf("%s", qPrintable(FormatMoney(actual)));
            }
            else
            {
                str.sprintf("%s", qPrintable(FormatMoney(actual)));
            }
            returnVal.setValue(str);
            return returnVal;
        }
        // Planned Column
        if(index.column() % NUM_COLUMNS == PLANNED_COLUMN)
        {
            QString str;
            if(pAccount->getType() == ACCOUNT_TYPE_TOTAL_INCOME)
            {

            }
            else if(pAccount->getType() == ACCOUNT_TYPE_TOTAL_EXPENSE)
            {
            }
            else if(pAccount->getType() == ACCOUNT_TYPE_INCOME)
            {
            }
            else
            {
                str.sprintf("%s", qPrintable(FormatMoney(planned)));
            }
            returnVal.setValue(str);
            return returnVal;
        }

        if(index.column() % NUM_COLUMNS == ACCUMULATED_COLUMN)
        {
            float actual, planned, accumulated;
            actual = 0;
            planned = 0;
            accumulated = 0;
            QVariant returnVal;
            // Total Value
            pBudget->ComputeMonthlyTotal(val.toString(), pBudget->columnToDate(index.column()), actual, planned, accumulated);


            AccountClass *pAccount = pBudget->getAccount(guid);

            QString str;
            if(pAccount->getType() == ACCOUNT_TYPE_TOTAL_INCOME)
            {
                /*
                if(planned != 0)
                {
                    str.sprintf("(+%s) %s", qPrintable(FormatMoney(planned)), qPrintable(FormatMoney(accumulated)));
                }
                else
                {
                    str.sprintf("%s", qPrintable(FormatMoney(accumulated)));
                }
                */
            }
            else if(pAccount->getType() == ACCOUNT_TYPE_TOTAL_EXPENSE)
            {
                //str.sprintf("%s", qPrintable(FormatMoney(accumulated)));
            }
            else if(pAccount->getType() == ACCOUNT_TYPE_INCOME)
            {
                //str.sprintf("%s", qPrintable(FormatMoney(actual)));
            }
            else
            {
                str.sprintf("%s", qPrintable(FormatMoney(accumulated)));
            }
            returnVal.setValue(str);
            return returnVal;
        }
    }
    return QVariant();
}

Qt::ItemFlags BudgetModel::flags ( const QModelIndex & index ) const
{
    if (!index.isValid())
        return Qt::NoItemFlags;

    Qt::ItemFlags flags = Qt::ItemIsSelectable | Qt::ItemIsEnabled;

    if(index.row() != 0
       && index.column() % NUM_COLUMNS == RULE_COLUMN)
        flags |= Qt::ItemIsEditable;

    return flags;
}

bool BudgetModel::setData(const QModelIndex &index, const QVariant &/*value*/, int /*role*/)
{
    if (!index.isValid() || (flags(index) & Qt::ItemIsEditable) == 0)
        return false;

    emit(dataChanged(index, index));
    /*
    switch (role)
    {
    case Qt::EditRole:
    case Qt::DisplayRole:
        {
            if(index.column() % 2 == 0)
            {
                pBudget->SetRule(value.
            }
        }
    }
    */

    return false;
}
