/*
  The code in this file is licensed under Public Domain.
  The license can be found at http://creativecommons.org/licenses/publicdomain/.
  */

#include "BudgetViewDelegate.h"

#include "RuleClass.h"
#include "RuleEdit.h"
#include "BudgetModel.h"
#include "Utilities.h"

#include <QPainter>
#include <QFontMetrics>
#include <iostream>


const QRgb ColorIncomeMismatch(0xffFF98CA);
const QRgb ColorIncomeTotal(0xff1FA48);
const QRgb ColorIncome(0xff1FACA);
const QRgb ColorExpenseTotal(0xff96EF96);
const QRgb ColorExpense(0xffCBEFCB);
const QRgb ColorExpenseAccumOverage(ColorIncomeMismatch);

BudgetViewDelegate::BudgetViewDelegate(QObject *pParent) : QStyledItemDelegate(pParent)
{
}

void BudgetViewDelegate::paint(QPainter *pPainter,
                               const QStyleOptionViewItem &option,
                               const QModelIndex &index) const
{
    QRect ruleRect = option.rect;
    if(index.column() % NUM_COLUMNS == RULE_COLUMN)
    {
        // Make room for the thick monthly seperator line
        ruleRect.adjust(3,0,0,0);
    }
    // Row 0 can use the regular painter
    if(index.row() == 0)
    {
        QStyledItemDelegate::paint(pPainter, option, index);
    }
    else
    {
        // Determine the account type
        QVariant val = index.model()->headerData(index.row(), Qt::Vertical, Qt::UserRole);
        QString guid = val.toString();
        const BudgetModel *pModel = dynamic_cast<const BudgetModel*>(index.model());
        AccountClass *pAccount = pModel->getAccount(index.row());
        Q_ASSERT(pAccount != NULL);

        // Draw the Rule for the expenses
        if(   pAccount->getType() == ACCOUNT_TYPE_EXPENSE
           && index.column() % NUM_COLUMNS == RULE_COLUMN)
        {
            QBrush brush;
            brush.setColor(QColor(ColorExpense));
            brush.setStyle(Qt::SolidPattern);
            pPainter->setBrush(brush);
            pPainter->drawRect(option.rect);

            // Get the rule
            RuleClass *pRule = index.model()->data(index, Qt::DisplayRole).value<RuleClass*>();

            // If no rule is specified, draw a line through the box
            if(pRule == NULL
               || ( pRule->getAccountMax() == 0.0
                   && pRule->getPercent() == 0.0
                   && pRule->getMonthlyFixed() == 0.0) )
            {
                int center = ruleRect.bottom() - ruleRect.height()/2;
                int inset = ruleRect.width()*2/3;
                QPoint p1(ruleRect.left()+inset, center);
                QPoint p2(ruleRect.right()-inset, center);

                pPainter->drawLine(p1, p2);
            }
            else
            {
                // Get the Monthly and Account Max strings
                QString monthlyString = pRule->getMontlyString();
                QString accountString = pRule->getAccountString();

                QFontMetrics metrics(option.font);
                QRect accountRect = metrics.boundingRect(accountString);

                if(!(pRule->getMonthlyFixed() == 0.0 && pRule->getPercent() == 0.0))
                {
                    // Draw the monthly text along the bottom of the cell
                    pPainter->drawText(ruleRect, monthlyString, Qt::AlignBottom | Qt::AlignCenter);
                }

                // Only draw the account max if it exists
                if(pRule->getAccountMax() != 0.0)
                {
                    // Draw the account max text along the top of the cell
                    QRect rect = ruleRect;
                    rect.adjust(0,-2*accountRect.height()/3,0,0);
                    pPainter->drawText(rect, accountString, Qt::AlignTop | Qt::AlignCenter);


                    rect = QRect(QPoint(ruleRect.width()/2 - accountRect.width()/2 + ruleRect.x(),
                                        ruleRect.top() + accountRect.height()/3),
                                 accountRect.size());
                    rect.adjust(-3,-1,3,-1);
                    pPainter->drawLine(rect.topLeft(), rect.topRight());
                    pPainter->drawLine(rect.bottomLeft(), rect.topLeft());
                    pPainter->drawLine(rect.topRight(), rect.bottomRight());
                }
            }
        }
        else
        {
            // Paint all other cells
            QString str;
            QColor bgColor(QRgb(0xff0000FF));
            if(pAccount->getType() == ACCOUNT_TYPE_TOTAL_INCOME
               && index.column() % NUM_COLUMNS == PLANNED_COLUMN
               && index.model()->data(index, Qt::DisplayRole).value<QString>() != QString())
            {
                bgColor = ColorIncomeMismatch;
                str = index.model()->data(index, Qt::DisplayRole).value<QString>();
            }
            else if(pAccount->getType() == ACCOUNT_TYPE_TOTAL_INCOME)
            {
                bgColor = ColorIncomeTotal;
                if(index.column() % NUM_COLUMNS == ACTUAL_COLUMN)
                    str = index.model()->data(index, Qt::DisplayRole).value<QString>();
            }
            else if(pAccount->getType() == ACCOUNT_TYPE_INCOME)
            {
                if(index.column() % NUM_COLUMNS != RULE_COLUMN)
                {
                    // Get the string
                    str = index.model()->data(index, Qt::DisplayRole).value<QString>();
                }
                bgColor = QRgb(0xff1FACA);
                bgColor = ColorIncome;
            }
            else if(pAccount->getType() == ACCOUNT_TYPE_TOTAL_EXPENSE)
            {
                bgColor = QRgb(0xff96EF96);
                bgColor = ColorExpenseTotal;
                if(index.column() % NUM_COLUMNS == ACTUAL_COLUMN
                   || index.column() % NUM_COLUMNS == PLANNED_COLUMN)
                {
                    str = index.model()->data(index, Qt::DisplayRole).value<QString>();
                }
            }
            else if(pAccount->getType() == ACCOUNT_TYPE_EXPENSE)
            {
                if(   index.column() % NUM_COLUMNS == ACCUMULATED_COLUMN
                   && pAccount->getCalculatedAccumulated(pModel->getDate(index.column())) < 0)
                    bgColor = ColorExpenseAccumOverage;
                else
                    bgColor = ColorExpense;//QRgb(0xffCBEFCB);
                str = index.model()->data(index, Qt::DisplayRole).value<QString>();
            }
            QBrush brush;
            brush.setColor(bgColor);
            brush.setStyle(Qt::SolidPattern);
            pPainter->setBrush(brush);
            pPainter->drawRect(option.rect);
            pPainter->drawText(option.rect, Qt::AlignCenter | Qt::AlignHCenter, str);
        }
    }

    if(index.column() % NUM_COLUMNS == RULE_COLUMN)
    {
        pPainter->save();
        // Draw an extra line against the left side of the Rule column
        QPen pen(QBrush(Qt::black), 3);
        pPainter->setPen(pen);
        QPoint p1, p2;
        p1.setX(option.rect.left());
        p1.setY(option.rect.top());
        p2.setX(option.rect.left());
        p2.setY(option.rect.bottom());
        pPainter->drawLine(p1,p2);
        pPainter->restore();
    }

}

QSize BudgetViewDelegate::sizeHint(const QStyleOptionViewItem &option,
                                   const QModelIndex &index) const
{
    QFont font("Times New Roman", 12);
    QFontInfo info(font);

    if(index.column() % NUM_COLUMNS == RULE_COLUMN)
    {
        // Get the rule
        RuleClass *pRule = index.model()->data(index, Qt::DisplayRole).value<RuleClass*>();
        if(!pRule)
        {
            return QStyledItemDelegate::sizeHint(option, index);
        }

        QString monthlyString = pRule->getMontlyString();
        QString accountString = pRule->getAccountString();

        QFontMetrics metrics(font);
        QSize monthlySize = metrics.boundingRect(monthlyString).size();
        QSize accountSize = metrics.boundingRect(accountString).size();

        QSize size;

        if(pRule->getAccountMax() == 0)
        {
            size = monthlySize;
        }
        else
        {
            // Add 8 pixels to accountSize 3 for each size + 1 to see the box
            size = QSize(qMax(monthlySize.width()+8,accountSize.width()+8),
                         monthlySize.height() + 4*accountSize.height()/3);
        }

        // Add a bit to the left side of the rule column for the month marker
        size.setWidth(size.width()+3);

        return size;
    }
    else if((index.column() % NUM_COLUMNS == ACTUAL_COLUMN
             || index.column() % NUM_COLUMNS == ACCUMULATED_COLUMN
             || index.column() % NUM_COLUMNS == PLANNED_COLUMN)
            && index.row() != 0 )
    {
        QString str = index.model()->data(index, Qt::DisplayRole).value<QString>();
        QFontMetrics metrics(font);
        QSize size = metrics.boundingRect(str).size();
        return size;

    }

    return QStyledItemDelegate::sizeHint(option, index);
}


QWidget *BudgetViewDelegate::createEditor(QWidget *parent,
                                   const QStyleOptionViewItem &option,
                                   const QModelIndex &index) const
{
    if(index.column() % 2 == 0)
    {
        // If on a mac, place a breakpoint here, then step up the stack a bit.
        // This allows the dialog to be seen.  I am not certain if this is an
        // unknown bug or a programming error.
        RuleEdit *pEditor = new RuleEdit(parent);
        return pEditor;
    }

    return QStyledItemDelegate::createEditor(parent, option, index);
}

void BudgetViewDelegate::setEditorData(QWidget *editor,
                                  const QModelIndex &index) const
 {
     if(index.column() % 2 == 0)
     {
         RuleEdit *pRuleEdit = qobject_cast<RuleEdit *>(editor);
         RuleClass *pRule = index.model()->data(index, Qt::DisplayRole).value<RuleClass*>();
         pRuleEdit->SetRule(pRule);
     }
     else
     {
         QStyledItemDelegate::setEditorData(editor, index);
     }
 }

void BudgetViewDelegate::setModelData(QWidget *editor, QAbstractItemModel *model,
                                 const QModelIndex &index) const
 {
     if(index.column() % 2 == 0)
     {
         RuleEdit *pRuleEdit = qobject_cast<RuleEdit *>(editor);
         RuleClass *pRule = index.model()->data(index, Qt::DisplayRole).value<RuleClass*>();
         pRuleEdit->GetRule(pRule);
         model->setData(index, QVariant());
     }
     else
     {
         QStyledItemDelegate::setModelData(editor, model, index);
     }
 }
