/*
  The code in this file is licensed under Public Domain.
  The license can be found at http://creativecommons.org/licenses/publicdomain/.
  */

#include "RuleClass.h"

#include "Logger.h"
#include "Utilities.h"

#include <QStringList>
#include <iostream>
#include <QDateTime>
#include <QUuid>

RuleClass::RuleClass(QString guid, QDate inputMonth)
{
    this->guid = guid;

    this->month = QDate(inputMonth.year(), inputMonth.month(), 1);

    accountMax = 0;
    monthlyMax = 0;
    monthlyFixed = 0;
    percent = 0;

    // Lookup the rule in the database
    QDateTime time = QDateTime();
    time.setDate(month);

    // Query the database for a rule
    QSqlQuery query(QString("SELECT * FROM budget_rules WHERE account_guid='%1' "
                            "AND month='%2';")
                    .arg(guid)
                    .arg(time.toString(TimeDateFormat)));
    SqlQueryVerify(query, "Error looking up budget rule");
    if(query.next())
    {

        accountMax = query.value(query.record().indexOf("accountMax")).toString().toFloat();
        monthlyMax = query.value(query.record().indexOf("monthlyMax")).toString().toFloat();
        monthlyFixed = query.value(query.record().indexOf("monthlyFixed")).toString().toFloat();
        percent = query.value(query.record().indexOf("percentage")).toString().toFloat();
    }

}

RuleClass::RuleClass()
{
}

QString RuleClass::getMontlyString(void)
{
    QString str;
    if(percent == 0)
    {
        return QString("$%1").arg(ConvertToString(monthlyFixed));
    }

    if(monthlyMax == 0.0)
    {
        if(monthlyFixed == 0.0)
            return QString("%2%").arg(ConvertToString(percent));
        else
            return QString("%1*%2%").arg(ConvertToString(monthlyFixed)).arg(ConvertToString(percent));
    }

    return QString("%1*%2%<$%3")
            .arg(ConvertToString(monthlyFixed))
            .arg(ConvertToString(percent))
            .arg(ConvertToString(monthlyMax));
}

QString RuleClass::getAccountString(void)
{
    return QString("$%1").arg(ConvertToString(accountMax));
}

QString RuleClass::ConvertToString(float value)
{
    QString str;
    if( (value - (int)value)*100 > 0)
        str.sprintf("%0.2f", value);
    else
        str.sprintf("%0.0f", value);
    return str;
}

float RuleClass::getAccountMax()
{
    return accountMax;
}

float RuleClass::getMonthlyMax()
{
    return monthlyMax;
}

float RuleClass::getMonthlyFixed()
{
    return monthlyFixed;
}

float RuleClass::getPercent()
{
    return percent;
}

QDate RuleClass::getDate(void)
{
    return month;
}

void RuleClass::setAccountMax(float val)
{
    if(accountMax != val)
    {
        accountMax = val;
        UpdateQuery();
    }
}

void RuleClass::setMonthlyMax(float val)
{
    if(monthlyMax != val)
    {
        monthlyMax = val;
        UpdateQuery();
    }
}

void RuleClass::setMonthlyFixed(float val)
{
    if(monthlyFixed != val)
    {
        monthlyFixed = val;
        UpdateQuery();
    }
}

void RuleClass::setPercent(float val)
{
    if(percent != val)
    {
        percent = val;
        UpdateQuery();
    }
}

void RuleClass::UpdateQuery(void)
{
    QDateTime datetime;
    datetime.setTime(QTime());
    datetime.setDate(month);

    bool emptyRule = false;
    if(accountMax == 0.0
       && monthlyFixed == 0.0
       && monthlyMax == 0.0
       && percent == 0.0)
    {
        emptyRule = true;
    }

    // Does the rule exist in the database yet?
    QSqlQuery query;
    query.exec(QString("SELECT * from budget_rules WHERE account_guid='%1' AND month='%2';")
               .arg(guid).arg(datetime.toString(TimeDateFormat)));
    if(query.next())
    {
        // If there are no values and it exists in the database, delete the rule
        if(emptyRule == true)
        {
            LogMsg("RuleClass/updte", LOG_DEBUG,
                   "Account:%s rule is empty. Deleting.",
                   qPrintable(guid));
            query.exec(QString("DELETE FROM budget_rules WHERE "
                               "account_guid='%1' AND month='%2';")
                       .arg(guid)
                       .arg(datetime.toString(TimeDateFormat)));
            SqlQueryVerify(query, "Error trying to delete rule.");
        }
        else
        {
            LogMsg("RuleClass/updte", LOG_DEBUG,
                   "Account:%s Updating rule.",
                   qPrintable(guid));
            // The rule exists.  Update it
            query.exec(QString("UPDATE budget_rules SET accountMax='%1', "
                               "monthlyMax='%2', monthlyFixed='%3', percentage='%4'"
                               "WHERE account_guid='%5' AND month='%6';")
                       .arg(accountMax)
                       .arg(monthlyMax)
                       .arg(monthlyFixed)
                       .arg(percent)
                       .arg(guid)
                       .arg(datetime.toString(TimeDateFormat)));
            SqlQueryVerify(query, "Error updating value.");
        }
    }
    else if(emptyRule == false)
    {
        LogMsg("RuleClass/updte", LOG_DEBUG,
               "Account:%s Creating new rule.",
               qPrintable(guid));
        // Create a new guid to accompany the rule
        QUuid newGuid = QUuid::createUuid();
        QString newGuidStr = newGuid.toString();
        newGuidStr = newGuidStr.remove("-");
        newGuidStr = newGuidStr.remove("{");
        newGuidStr = newGuidStr.remove("}");
        // The rule doesn't exist yet.  Add it.
        query.prepare(QString("INSERT INTO budget_rules (guid, account_guid, month, accountMax, monthlyMax, monthlyFixed, percentage) "
                      "VALUES ('%1', '%2', '%3', '%4', '%5', '%6', '%7');")
                      .arg(newGuidStr)
                      .arg(guid)
                      .arg(datetime.toString(TimeDateFormat))
                       .arg(accountMax)
                       .arg(monthlyMax)
                       .arg(monthlyFixed)
                       .arg(percent));
        query.exec();
        SqlQueryVerify(query, "Error inserting Rule into the database.");
    }
}
